# index and level name
indexes_schema_level_map = {
    "files": "dataset",
    "gct_metadata": "sample",
    "h5ad_metadata": "sample",
}

schema_levels = ["dataset", "sample"]

schema_table_names = [
    "datasets",
    "samples",
    "features",
    "samples_singlecell",
    "features_singlecell",
    "variant_data",
]
